"""
Module containing functions to compute orbital properties from initial conditions
"""

from .frequencies_from_constants import r_frequency_from_constants, theta_frequency_from_constants
from .plunging_solutions import _plunging_radial_roots, plunging_mino_frequencies
import numpy as np
from numpy import sin, cos, sqrt, pi, arcsin, arccos
from numpy.polynomial import Polynomial
from scipy.special import ellipkinc

def apex_from_constants(a,E,L,Q):
    r"""
    Computes the orbital parameters :math:`(a,p,e,x)` for a stable bound orbit with the given constants of motion
    
    :param a: spin parameter
    :type a: double
    :param E: dimensionless energy
    :type E: double
    :param L: dimensionless angular momentum
    :type L: double
    :param Q: dimensionless Carter constant
    :type Q: double

    :return: tuple of orbital parameters :math:`(a,p,e,x)`
    :rtype: tuple(double,double,double,double)
    """
    # radial polynomial written in terms of z = cos^2(theta)
    R = Polynomial([-a**2*Q, 2*L**2+2*Q+2*a**2*E**2-4*a*E*L, a**2*E**2-L**2-Q-a**2, 2, E**2-1])
    radial_roots = R.roots()
    # numpy returns roots in increasing order
    r4, r3, r2, r1 = radial_roots

    # polar polynomial written in terms of z = cos^2(theta)
    Z = Polynomial([Q,-(Q+a**2*(1-E**2)+L**2),a**2*(1-E**2)])
    polar_roots = Z.roots()
    z_minus, z_plus = polar_roots

    p = 2*r1*r2/(r1+r2)
    e = (r1-r2)/(r1+r2)
    x = np.sign(L)*sqrt(1-z_minus)

    return a, p, e, x

def constants_from_initial_conditions(a,initial_position,initial_velocity):
    r"""
    Computes the constants of motion :math:`(E,L,Q)` of the orbit with the given initial conditions

    :param a: spin parameter
    :type a: double
    :param initial_position: initial position :math:`(t_0,r_0,\theta_0,\phi_0)`
    :type initial_position: tuple(double,double,double,double)
    :param initial_velocity: initial four-velocity :math:`(u^t_0,u^r_0,u^\theta_0,u^\phi_0)`
    :type initial_velocity: tuple(double,double,double,double)

    :return: tuple of constants of motion :math:`(E,L,Q)`
    :rtype: tuple(double,double,double)
    """

    t0, r0, theta0, phi0 = initial_position
    dt0, dr0, dtheta0, dphi0 = initial_velocity

    sigma = r0**2+a**2*cos(theta0)**2
    delta = r0**2-2*r0+a**2

    # solve for E and L by writing the t and phi geodesic equations as a matrix equation
    A = np.array(
        [[(r0**2+a**2)**2/delta-a**2*(1-cos(theta0)**2), a-a*(r0**2+a**2)/delta],
        [a*(r0**2+a**2)/delta-a, 1/(1-cos(theta0)**2)-a**2/delta]]
    )
    b = np.array([sigma*dt0,sigma*dphi0])
    E, L = np.linalg.solve(A,b)

    # solve for Q by substituting E and L back into the theta equation
    Q = ((sigma*dtheta0*sin(theta0))**2 + cos(theta0)**2*(a**2*(1-E**2)*(1-cos(theta0)**2)+L**2))/(1-cos(theta0)**2)

    return E, L, Q

def is_stable(a,initial_position,initial_velocity,constants=None,tol=1e-4):
    r"""
    Tests whether the orbit with the given initial conditions is stable

    :param a: spin parameter
    :type a: double
    :param initial_position: initial position :math:`(t_0,r_0,\theta_0,\phi_0)`
    :type initial_position: tuple(double,double,double,double)
    :param initial_velocity: initial four-velocity :math:`(u^t_0,u^r_0,u^\theta_0,u^\phi_0)`
    :type initial_velocity: tuple(double,double,double,double)
    :param constants: tuple of constants of motion :math:`(E,L,Q)` can be passed in to avoid recomputing them
    :type constants: tuple(double,double,double), optional
    :param tol: numerical tolerance used when comparing :math:`r_0` to radial roots, defaults to 1e-4
    :type tol: double, optional

    :return: True if the orbit is stable, False otherwise
    :rtype: bool
    """

    t0, r0, theta0, phi0 = initial_position
    # recompute constants if they are not passed in
    if constants is None: constants = constants_from_initial_conditions(a,initial_position,initial_velocity)
    E, L, Q = constants

    # radial polynomial
    R = Polynomial([-a**2*Q, 2*L**2+2*Q+2*a**2*E**2-4*a*E*L, a**2*E**2-L**2-Q-a**2, 2, E**2-1])
    radial_roots = R.roots()

    # get the real roots and the complex roots
    real_roots = np.sort(np.real(radial_roots[np.isreal(radial_roots)]))
    complex_roots = radial_roots[np.iscomplex(radial_roots)]

    # event horizon radius
    r_minus = 1-sqrt(1-a**2)

    if len(complex_roots) == 4: 
        raise ValueError("Not a physical orbit")
    
    if len(complex_roots) == 2:
        return False
    
    r4, r3, r2, r1 = real_roots
    # r4 < r3 < [r2 < r_minus < r_plus < r1]
    if (r2 < r_minus) & (r0 <= r1+tol) & (r0 >= r2-tol):
        return False
    # r4 < r_minus < r_plus < r3 < [r2 < r1]
    if (r2 > r_minus) & (r0 <= r1+tol) & (r0 >= r2-tol):
        return True
    # [r4 < r_minus < r_plus < r3] < r2 < r1
    if (r2 > r_minus) & (r0 <= r3+tol) & (r0 >= r4-tol):
        return False
    
    raise ValueError("Not a physical orbit")
    

def stable_orbit_initial_phases(a,initial_position,initial_velocity,constants=None,radial_roots=None,upsilon_r=None,upsilon_theta=None,tol=1e-4):
    r"""
    Computes the initial phases :math:`(q_{t_0},q_{r_0},q_{\theta_0},q_{\phi_0})` of a stable bound orbit with the given initial conditions.
    Computes phases with respect to the starting point :math:`(t_0,r_0,\theta_0,\phi_0) = (0,r_\text{min},\theta_\text{min},0)`.

    :param a: spin parameter
    :type a: double
    :param initial_position: initial position :math:`(t_0,r_0,\theta_0,\phi_0)`
    :type initial_position: tuple(double,double,double,double)
    :param initial_velocity: initial four-velocity :math:`(u^t_0,u^r_0,u^\theta_0,u^\phi_0)`
    :type initial_velocity: tuple(double,double,double,double)
    :param constants: constants of motion :math:`(E,L,Q)` can be passed in to avoid recomputing them
    :type constants: tuple(double,double,double), optional
    :param radial_roots: radial roots :math:`(r_1,r_2,r_3,r_4)` can be passed in to avoid recomputing them
    :type radial_roots: tuple(double,double,double,double), optional
    :param tol: numerical tolerance used when checking for turning points, defaults to 1e-4
    :type tol: double, optional

    :return: tuple of initial phases :math:`(q_{t_0},q_{r_0},q_{\theta_0},q_{\phi_0})`
    :rtype: tuple(double,double,double,double)
    """
    t0, r0, theta0, phi0 = initial_position
    dt0, dr0, dtheta0, dphi0 = initial_velocity

    # recompute constants if they are not passed in
    if constants is None: constants = constants_from_initial_conditions(a,initial_position,initial_velocity)
    E, L, Q = constants

    # polar polynomial written in terms of z = cos^2(theta)
    Z = Polynomial([Q,-(Q+a**2*(1-E**2)+L**2),a**2*(1-E**2)])
    polar_roots = Z.roots()
    z_minus, z_plus = polar_roots

    # recompute radial roots if they are not passed in
    if radial_roots is None:
        R = Polynomial([-a**2*Q, 2*L**2+2*Q+2*a**2*E**2-4*a*E*L, a**2*E**2-L**2-Q-a**2, 2, E**2-1])
        radial_roots = R.roots()
        # numpy returns roots in increasing order
        r4, r3, r2, r1 = radial_roots
    else:
        r1, r2, r3, r4 = radial_roots

    # check for turning points in r
    if abs(r0-r2) < tol:
        q_r0 = 0
    elif abs(r0-r1) < tol:
        q_r0 = pi
    else:
        # compute r frequency
        if upsilon_r is None: upsilon_r = r_frequency_from_constants(constants,radial_roots)

        # Fujita and Hikida equation 13
        k_r = sqrt((r1-r2)*(r3-r4)/((r1-r3)*(r2-r4)))
        y_r = sqrt((r1-r3)*(r0-r2)/((r1-r2)*(r0-r3)))
        # Fujita and Hikida equation 26
        lambda_r0 = 1/sqrt(1-E**2)*2/sqrt((r1-r3)*(r2-r4))*ellipkinc(arcsin(y_r),k_r**2)

        # Fujita and Hikida equation 24
        q_r0 = lambda_r0*upsilon_r if dr0 > 0 else 2*pi-lambda_r0*upsilon_r

    # Fujita and Hikida equation 11
    theta_min = arccos(sqrt(z_minus))
    theta_max = pi-theta_min

    # check for turning points in theta
    if abs(theta0-theta_min) < tol:
        q_theta0 = 0
    elif abs(theta0-theta_max) < tol:
        q_theta0 = pi
    else:
        # compute theta frequency
        if upsilon_theta is None: upsilon_theta = theta_frequency_from_constants(a,constants,radial_roots,polar_roots)

        k_theta = sqrt(z_minus/z_plus)
        y_theta = cos(theta0)/sqrt(z_minus)
        e0zp = (a**2*(1-E**2)*(1-z_minus)+L**2)/(L**2*(1-z_minus))

        # Fujita and Hikida equation 37
        lambda_theta0 = 1/(L*sqrt(e0zp))*ellipkinc(arcsin(y_theta),k_theta**2)

        # Fujita and Hikida equation 36
        if dtheta0 < 0: q_theta0 = lambda_theta0*upsilon_theta+3*pi/2
        if (dtheta0 == 0) & (theta0 == pi/2): q_theta0 = 0 # special case for equatorial orbits
        if dtheta0 >= 0: q_theta0 = pi/2-lambda_theta0*upsilon_theta

    q_t0 = t0
    q_phi0 = phi0

    return q_t0, q_r0, q_theta0, q_phi0

def plunging_orbit_initial_phases(a,initial_position,initial_velocity,constants=None,radial_roots=None,tol=1e-4):
    r"""
    Computes the initial phases :math:`(q_{t_0},q_{r_0},q_{\theta_0},q_{\phi_0})` of a plunging orbit with the given initial conditions.
    Computes phases with respect to the starting point :math:`(t_0,r_0,\theta_0,\phi_0) = (0,r_\text{min},\theta_{\text{min}},0)`.

    :param a: spin parameter
    :type a: double
    :param initial_position: initial position :math:`(t_0,r_0,\theta_0,\phi_0)`
    :type initial_position: tuple(double,double,double,double)
    :param initial_velocity: initial four-velocity :math:`(u^t_0,u^r_0,u^\theta_0,u^\phi_0)`
    :type initial_velocity: tuple(double,double,double,double)
    :param constants: constants of motion :math:`(E,L,Q)` can be passed in to avoid recomputing them
    :type constants: tuple(double,double,double), optional
    :param radial_roots: radial roots :math:`(r_1,r_2,r_3,r_4)` can be passed in to avoid recomputing them
    :type radial_roots: tuple(double,double,double,double), optional
    :param tol: numerical tolerance used when checking for turning points, defaults to 1e-4
    :type tol: double, optional

    :return: tuple of initial phases :math:`(q_{t_0},q_{r_0},q_{\theta_0},q_{\phi_0})`
    :rtype: tuple(double,double,double,double)
    """
    t0, r0, theta0, phi0 = initial_position
    dt0, dr0, dtheta0, dphi0 = initial_velocity

    # recompute constants if they are not passed in
    if constants is None: constants = constants_from_initial_conditions(a,initial_position,initial_velocity)
    E, L, Q = constants

    # recompute radial roots if they are not passed in
    if radial_roots is None: radial_roots = _plunging_radial_roots(a,E,L,Q)
    r1, r2, r3, r4 = radial_roots


    if np.iscomplex(r3):
        z1 = sqrt(1/2*(1+(L**2+Q)/(a**2*(1-E**2))-sqrt((1+(L**2+Q)/(a**2*(1-E**2)))**2-4*Q/(a**2*(1-E**2)))))
        z2 = sqrt(a**2*(1-E**2)/2*(1+(L**2+Q)/(a**2*(1-E**2))+sqrt((1+(L**2+Q)/(a**2*(1-E**2)))**2-4*Q/(a**2*(1-E**2)))))
        k_theta = a*sqrt(1-E**2)*z1/z2
        rho_r = np.real(r3)
        rho_i = np.imag(r4)

        # equation 47
        A = sqrt((r2-rho_r)**2+rho_i**2)
        B = sqrt((r1-rho_r)**2+rho_i**2)
        k_r = sqrt(((r2-r1)**2-(A-B)**2)/(4*A*B))
        # equation 49
        lambda_r0 = -1/sqrt((1-E**2)*A*B)*ellipkinc(pi/2-arcsin((B*(r2-r0)-A*(r0-r1))/(B*(r2-r0)+A*(r0-r1))),k_r**2)
        # derived by inverting equation 29
        lambda_theta0 = 1/z2*ellipkinc(arcsin(cos(theta0)/z1),k_theta**2)

        upsilon_r, upsilon_theta = plunging_mino_frequencies(a,E,L,Q)

        q_t0 = t0
        q_r0 = -lambda_r0*upsilon_r if dr0 > 0 else 2*pi+lambda_r0*upsilon_r

        if dtheta0 < 0: q_theta0 = lambda_theta0*upsilon_theta+3*pi/2
        if (dtheta0 == 0) & (theta0 == pi/2): q_theta0 = 0 # special case for equatorial orbits
        if dtheta0 >= 0: q_theta0 = pi/2-lambda_theta0*upsilon_theta

        q_phi0 = phi0

        return q_t0, q_r0, q_theta0, q_phi0
    else:
        return stable_orbit_initial_phases(a,initial_position,initial_velocity,constants=constants,radial_roots=radial_roots,tol=tol)


"""
Module containing the PlungingOrbit class
"""
from .plunging_solutions import *
from .plunging_solutions import _plunging_radial_roots
from .stable_solutions import *
from .orbit import Orbit

class PlungingOrbit(Orbit):
    r"""
    Class representing a plunging orbit in Kerr spacetime.

    :param a: dimensionaless spin parameter
    :type a: double
    :param E: dimensionless energy
    :type E: double
    :param L: dimensionless angular momentum
    :type L: double
    :param Q: dimensionless Carter constant
    :type Q: double
    :param initial_phases: tuple of initial phases :math:`(q^t_0, q^r_0, q^\theta_0, q^\phi_0)`, defaults to (0,0,0,0)
    :type initial_phases: tuple, optional
    :param M: mass of the primary in solar masses, optional
    :type M: double
    :param mu: mass of the smaller body in solar masses, optional
    :type mu: double

    :ivar a: dimensionaless spin parameter
    :ivar E: dimensionless energy
    :ivar L: dimensionless angular momentum
    :ivar Q: dimensionless Carter constant
    :ivar initial_position: tuple of initial position coordinates :math:`(t_0, r_0, \theta_0, \phi_0)`
    :ivar initial_velocity: tuple of initial four-velocity components :math:`(u^t_0, u^r_0, u^\theta_0, u^\phi_0)`
    :ivar M: mass of the primary in solar masses
    :ivar mu: mass of the smaller body in solar masses
    :ivar upsilon_r: radial Mino frequency
    :ivar upsilon_theta: polar Mino frequency
    """
    def __init__(self,a,E,L,Q,initial_phases=(0,0,0,0),M=None,mu=None):
        self.a, self.E, self.L, self.Q, self.initial_phases, self.M, self.mu = a, E, L, Q, initial_phases, M, mu
        self.upsilon_r, self.upsilon_theta = plunging_mino_frequencies(a,E,L,Q)
        self.stable = False

        u_t, u_r, u_theta, u_phi = self.four_velocity()
        t, r, theta, phi = self.trajectory()
        self.initial_position = t(0), r(0), theta(0), phi(0)
        self.initial_velocity = u_t(0), u_r(0), u_theta(0), u_phi(0)

    def trajectory(self,initial_phases=None,distance_units="natural",time_units="natural"):
        r"""
        Computes the components of the trajectory as a function of mino time

        :param initial_phases: tuple of initial phases :math:`(q^t_0, q^r_0, q^\theta_0, q^\phi_0)`
        :type initial_phases: tuple, optional
        :param distance_units: units to compute the radial component of the trajectory in (options are "natural", "mks", "cgs", "au" and "km"), defaults to "natural"
        :type distance_units: str, optional
        :param time_units: units to compute the time component of the trajectory in (options are "natural", "mks", "cgs", and "days"), defaults to "natural"
        :type time_units: str, optional
        
        :return: tuple of functions :math:`t(\lambda), r(\lambda), \theta(\lambda), \phi(\lambda)`
        :rtype: tuple(function,function,function,function)
        """
        
        if distance_units != "natural" and (self.M is None or self.mu is None): raise ValueError("M and mu must be specified to convert r to physical units")
        if time_units != "natural" and (self.M is None or self.mu is None): raise ValueError("M and mu must be specified to convert t to physical units")

        a, E, L, Q = self.a, self.E, self.L, self.Q
        radial_roots = _plunging_radial_roots(a,E,L,Q)

        if np.iscomplex(radial_roots[3]):
            return self._complex_trajectory(initial_phases,distance_units,time_units)
        else:
            return self._real_trajectory(initial_phases,distance_units,time_units)
    
    def _complex_trajectory(self,initial_phases=None, distance_units="natural",time_units="natural"):
        r"""
        Computes the components of the trajectory in the case of two complex and two real radial roots

        :param initial_phases: tuple of initial phases :math:`(q^t_0, q^r_0, q^\theta_0, q^\phi_0)`, defaults to (0,0,0,0)
        :type initial_phases: tuple, optional

        :rtype: tuple(function,function,function,function)
        """
        if initial_phases is None: initial_phases = self.initial_phases
        a, E, L, Q = self.a, self.E, self.L, self.Q
        upsilon_r, upsilon_theta = self.upsilon_r, self.upsilon_theta
        r_phases, t_r, phi_r = plunging_radial_solutions_complex(a,E,L,Q)
        theta_phases, t_theta, phi_theta = plunging_polar_solutions(a,E,L,Q)
        q_t0, q_r0, q_theta0, q_phi0 = initial_phases

        distance_conversion_func = {"natural": lambda x,M: x, "mks": distance_in_meters, "cgs": distance_in_cm, "au": distance_in_au,"km": distance_in_km}
        time_conversion_func = {"natural": lambda x,M: x, "mks": time_in_seconds, "cgs": time_in_seconds, "days": time_in_days}

        # adjust q_theta0 so that initial conditions are consistent with stable case
        q_theta0 = q_theta0 + pi/2

        # Calculate normalization constants so that t = 0 and phi = 0 at lambda = 0 when q_t0 = 0 and q_phi0 = 0 
        C_t = t_r(q_r0)+t_theta(q_theta0)
        C_phi= phi_r(q_r0)+phi_theta(q_theta0)


        def t(mino_time):
            return time_conversion_func[time_units](t_r(upsilon_r*mino_time+q_r0) + t_theta(upsilon_theta*mino_time+q_theta0) - C_t + q_t0, self.M)
        
        def r(mino_time):
            return distance_conversion_func[distance_units](r_phases(upsilon_r*mino_time+q_r0), self.M)
        
        def theta(mino_time):
            return theta_phases(upsilon_theta*mino_time+q_theta0)
        
        def phi(mino_time):
            return phi_r(upsilon_r*mino_time+q_r0) + phi_theta(upsilon_theta*mino_time+q_theta0) - C_phi + q_phi0
        
        return t, r, theta, phi
    
    def _real_trajectory(self,initial_phases=None,distance_units="natural",time_units="natural"):
        r"""
        Computes the components of the trajectory in the case of four real radial roots

        :param initial_phases: tuple of initial phases :math:`(q^t_0, q^r_0, q^\theta_0, q^\phi_0)`
        :type initial_phases: tuple, optional

        :rtype: tuple(function,function,function,function)
        """
        if initial_phases is None: initial_phases = self.initial_phases
        a, E, L, Q = self.a, self.E, self.L, self.Q
        constants = (E,L,Q)
        radial_roots = _plunging_radial_roots(a,E,L,Q)

        # polar polynomial written in terms of z = cos^2(theta)
        Z = Polynomial([Q,-(Q+a**2*(1-E**2)+L**2),a**2*(1-E**2)])
        polar_roots = Z.roots()

        upsilon_r, upsilon_theta, upsilon_phi, gamma = mino_frequencies_from_constants(a,constants,radial_roots,polar_roots)
        r_phases, t_r, phi_r = radial_solutions(a,constants,radial_roots)
        theta_phases, t_theta, phi_theta = polar_solutions(a,constants,polar_roots)
        q_t0, q_r0, q_theta0, q_phi0 = initial_phases

        distance_conversion_func = {"natural": lambda x,M: x, "mks": distance_in_meters, "cgs": distance_in_cm, "au": distance_in_au,"km": distance_in_km}
        time_conversion_func = {"natural": lambda x,M: x, "mks": time_in_seconds, "cgs": time_in_seconds, "days": time_in_days}

        # Calculate normalization constants so that t = 0 and phi = 0 at lambda = 0 when q_t0 = 0 and q_phi0 = 0 
        C_t = t_r(q_r0)+t_theta(q_theta0)
        C_phi= phi_r(q_r0)+phi_theta(q_theta0)

        def t(mino_time):
            # equation 6
            return time_conversion_func[time_units](q_t0 + gamma*mino_time + t_r(upsilon_r*mino_time+q_r0) + t_theta(upsilon_theta*mino_time+q_theta0) - C_t, self.M)
        
        def r(mino_time):
            return distance_conversion_func[distance_units](r_phases(upsilon_r*mino_time+q_r0), self.M)
        
        def theta(mino_time):
            return theta_phases(upsilon_theta*mino_time+q_theta0)
        
        def phi(mino_time):
            # equation 6
            return q_phi0 + upsilon_phi*mino_time + phi_r(upsilon_r*mino_time+q_r0) + phi_theta(upsilon_theta*mino_time+q_theta0) - C_phi
        
        return t, r, theta, phi


import unittest
import numpy as np
from kerrgeopy.frequencies import *
from pathlib import Path

THIS_DIR = Path(__file__).parent

DATA_DIR = THIS_DIR.parent / "tests/data"

class TestFrequencies(unittest.TestCase):
    def test_extreme_kerr(self):
        with self.assertRaises(ValueError): mino_frequencies(1,12,0.5,0.5)
        with self.assertRaises(ValueError): r_frequency(1,0.5,12,0.5)
        with self.assertRaises(ValueError): theta_frequency(1,0.5,12,0.5)
        with self.assertRaises(ValueError): phi_frequency(1,0.5,12,0.5)

    def test_polar(self):
        with self.assertRaises(ValueError): mino_frequencies(0.5,12,0.5,0)
        with self.assertRaises(ValueError): r_frequency(0.5,0.5,12,0)
        with self.assertRaises(ValueError): theta_frequency(0.5,0.5,12,0)
        with self.assertRaises(ValueError): phi_frequency(0.5,0.5,12,0)

    def test_marginally_bound(self):
        with self.assertRaises(ValueError): mino_frequencies(0.5,12,1,0.5)
        with self.assertRaises(ValueError): r_frequency(0.5,12,1,0.5)
        with self.assertRaises(ValueError): theta_frequency(0.5,12,1,0.5)
        with self.assertRaises(ValueError): phi_frequency(0.5,12,1,0.5)

    def test_invalid_arguments(self):
        with self.assertRaises(ValueError): mino_frequencies(2,5,0.5,0.5)
        with self.assertRaises(ValueError): r_frequency(0.5,5,-0.5,0.5)
        with self.assertRaises(ValueError): theta_frequency(0.5,5,0.5,-2)
        with self.assertRaises(ValueError): phi_frequency(0.5,5,2,0.5)

    def test_unstable(self):
        with self.assertRaises(ValueError): mino_frequencies(0.5,5,0.5,0.5)
        with self.assertRaises(ValueError): r_frequency(0.5,5,0.5,0.5)
        with self.assertRaises(ValueError): theta_frequency(0.5,5,0.5,0.5)
        with self.assertRaises(ValueError): phi_frequency(0.5,5,0.5,0.5)

    def test_random(self):
        values = np.genfromtxt(DATA_DIR / "freq_values.txt",delimiter=",")
        mathematica_freq_output = np.genfromtxt(DATA_DIR / "mathematica_freq_output.txt")
        python_freq_output = np.apply_along_axis(lambda x: mino_frequencies(*x),1,values)
        for i, params in enumerate(values):
            with self.subTest(i=i,params=params):
                E, L, Q = constants_of_motion(*params)
                self.assertTrue(np.allclose(abs(mathematica_freq_output[i]),abs(python_freq_output[i])))



"""
Module containing functions to compute intermediate terms in the stable bound orbit solutions of `Fujita and Hikida <https://doi.org/10.48550/arXiv.0906.1420>`_
"""
from .constants import *
from .frequencies_from_constants import _ellippi
from .frequencies_from_constants import *
from scipy.special import ellipj, ellipeinc
from numpy import pi, sin, cos, arcsin, arccos, floor, where

def _ellippiinc(phi,n,k):
    r"""
    Incomplete elliptic integral of the third kind defined as :math:`\Pi(\phi,n,k) = \int_0^{\phi} \frac{1}{1-n\sin^2\theta}\frac{1}{\sqrt{1-k^2\sin^2\theta}}d\theta`.

    :type phi: double
    :type n: double
    :type k: double

    :rtype: double
    """
    # Note: sign of n is reversed from the definition in Fujita and Hikida

    # count the number of half periods
    num_cycles = floor(phi/(pi/2))
    # map phi to [0,pi/2]
    phi = abs(arcsin(sin(phi)))
    # formula from https://en.wikipedia.org/wiki/Carlson_symmetric_form
    integral = sin(phi)*elliprf(cos(phi)**2,1-k**2*sin(phi)**2,1)+1/3*n*sin(phi)**3*elliprj(cos(phi)**2,1-k**2*sin(phi)**2,1,1-n*sin(phi)**2)

    return where(num_cycles % 2 == 0, num_cycles*_ellippi(n,k)+integral, (num_cycles+1)*_ellippi(n,k)-integral)
    
def radial_solutions(a,constants,radial_roots):
    r"""
    Computes the radial solutions :math:`r(q_r), t^{(r)}(q_r), \phi^{(r)}(q_r)` from equation 6 of `Fujita and Hikida <https://doi.org/10.48550/arXiv.0906.1420>`_.
    :math:`q_r` is defined as :math:`q_r = \Upsilon_r \lambda = 2\pi \frac{\lambda}{\Lambda_r}`.
    Assumes the initial conditions :math:`r(0) = r_{\text{min}}` and :math:`\theta(0) = \theta_{\text{min}}`.

    :param a: dimensionless spin parameter
    :type a: double
    :param constants: tuple of constants :math:`(E,L,Q)`
    :type constants: tuple(double,double,double)
    :param radial_roots: tuple of roots :math:`(r_1,r_2,r_3,r_4)`. Assumes that motion is between :math:`r_1` and :math:`r_2` and that roots are otherwise in decreasing order.
    :type radial_roots: tuple(double,double,double,double)

    :return: tuple of functions in the form :math:`(r, t^{(r)}, \phi^{(r)})`
    :rtype: tuple(function, function, function)
    """
    E, L, Q = constants
    r1, r2, r3, r4 = radial_roots

    r_plus = 1+sqrt(1-a**2)
    r_minus = 1-sqrt(1-a**2)
    
    h_r = (r1-r2)/(r1-r3)
    h_plus = (r1-r2)*(r3-r_plus)/((r1-r3)*(r2-r_plus))
    h_minus = (r1-r2)*(r3-r_minus)/((r1-r3)*(r2-r_minus))
    
    # equation 13
    k_r = sqrt((r1-r2)*(r3-r4)/((r1-r3)*(r2-r4)))

    def r(q_r):
        # equation 27
        u_r = ellipk(k_r**2)*q_r/pi

        sn, cn, dn, psi_r = ellipj(u_r,k_r**2)
        return (r3*(r1-r2)*sn**2-r2*(r1-r3))/((r1-r2)*sn**2-(r1-r3))
    
    def t_r(q_r):
        # equation 27
        u_r = ellipk(k_r**2)*q_r/pi
        sn, cn, dn, psi_r = ellipj(u_r,k_r**2)
        # equation 28
        return 2/sqrt((1-E**2)*(r1-r3)*(r2-r4))* \
        (
        E/2*(
            (r2-r3)*(r1+r2+r3+r4)*(_ellippiinc(psi_r,h_r,k_r)-q_r/pi*_ellippi(h_r,k_r)) \
            + (r1-r3)*(r2-r4)*(ellipeinc(psi_r,k_r**2)+h_r*sn*cn*sqrt(1-k_r**2*sn**2)/(h_r*sn**2-1) - q_r/pi*ellipe(k_r**2))
            ) 
        + 2*E*(r2-r3)*(_ellippiinc(psi_r,h_r,k_r)-q_r/pi*_ellippi(h_r,k_r)) 
        - 2/(r_plus-r_minus) * \
           (
            ((4*E-a*L)*r_plus-2*a**2*E)*(r2-r3)/((r3-r_plus)*(r2-r_plus))*(_ellippiinc(psi_r,h_plus,k_r)-q_r/pi*_ellippi(h_plus,k_r)) 
            - ((4*E-a*L)*r_minus-2*a**2*E)*(r2-r3)/((r3-r_minus)*(r2-r_minus))*(_ellippiinc(psi_r,h_minus,k_r)-q_r/pi*_ellippi(h_minus,k_r))
           )
        )
    
    def phi_r(q_r):
        # equation 27
        u_r = ellipk(k_r**2)*q_r/pi
        sn, cn, dn, psi_r = ellipj(u_r,k_r**2)
        # equation 28
        return -2*a/((r_plus-r_minus)*sqrt((1-E**2)*(r1-r3)*(r2-r4))) * \
                (
                (2*E*r_plus-a*L)*(r2-r3)/((r3-r_plus)*(r2-r_plus))*(_ellippiinc(psi_r,h_plus,k_r)-q_r/pi*_ellippi(h_plus,k_r)) 
                - (2*E*r_minus-a*L)*(r2-r3)/((r3-r_minus)*(r2-r_minus))*(_ellippiinc(psi_r,h_minus,k_r)-q_r/pi*_ellippi(h_minus,k_r))
                )
    
    return r, t_r, phi_r
        
def polar_solutions(a,constants,polar_roots):
    r"""
    Computes the polar solutions :math:`\theta(q_\theta), t^{(\theta)}(q_\theta), \phi^{(\theta)}(q_\theta)` from equation 6 of `Fujita and Hikida <https://doi.org/10.48550/arXiv.0906.1420>`_.
    :math:`q_\theta` is defined as :math:`q_\theta = \Upsilon_\theta \lambda = 2\pi \frac{\lambda}{\Lambda_\theta}`.
    Assumes the initial conditions :math:`r(0) = r_{\text{min}}` and :math:`\theta(0) = \theta_{\text{min}}`.

    :param a: dimensionless spin parameter
    :type a: double
    :param constants: tuple of constants :math:`(E,L,Q)`
    :type constants: tuple(double,double,double)
    :param polar_roots: tuple of roots :math:`(z_-,z_+)`
    :type polar_roots: tuple(double,double)

    :return: tuple of functions in the form :math:`(\theta, t^{(\theta)}, \phi^{(\theta)})`
    :rtype: tuple(function, function, function)
    """
    E, L, Q = constants
    z_minus, z_plus = polar_roots
    epsilon0 = a**2*(1-E**2)/L**2
    # simplified form of epsilon0*z_plus
    e0zp = (a**2*(1-E**2)*(1-z_minus)+L**2)/(L**2*(1-z_minus))
    # simplified form of a**2*sqrt(z_plus/epsilon0)
    a2sqrt_zp_over_e0 = L**2/((1-E**2)*sqrt(1-z_minus)) if a == 0 else a**2*z_plus/sqrt(epsilon0*z_plus)

    # equation 13
    k_theta = 0 if a == 0 else sqrt(z_minus/z_plus)

    def theta(q_theta):
        u_theta = 2/pi*ellipk(k_theta**2)*(q_theta+pi/2)
        sn, cn, dn, ph = ellipj(u_theta,k_theta**2)
        # equation 38
        return arccos(sqrt(z_minus)*sn)

    def t_theta(q_theta):
        u_theta = 2/pi*ellipk(k_theta**2)*(q_theta+pi/2)
        sn, cn, dn, psi_theta = ellipj(u_theta,k_theta**2)
        # equation 39
        return sign(L)*a2sqrt_zp_over_e0*E/L*(2/pi*ellipe(k_theta**2)*(q_theta+pi/2)-ellipeinc(psi_theta,k_theta**2))
    
    def phi_theta(q_theta):
        sn, cn, dn, psi_theta = ellipj(2/pi*ellipk(k_theta**2)*(q_theta+pi/2),k_theta**2)
        # equation 39
        return sign(L)*1/sqrt(e0zp)*(_ellippiinc(psi_theta,z_minus,k_theta)-2/pi*_ellippi(z_minus,k_theta)*(q_theta+pi/2))
    
    return theta, t_theta, phi_theta




import unittest
import numpy as np
from kerrgeopy.stable_orbit import *
from pathlib import Path

THIS_DIR = Path(__file__).parent

DATA_DIR = THIS_DIR.parent / "tests/data"

class TestOrbit(unittest.TestCase):
    def test_extreme_kerr(self):
        with self.assertRaises(ValueError): StableOrbit(1,12,0.5,0.5)

    def test_polar(self):
        with self.assertRaises(ValueError): StableOrbit(0.5,12,0.5,0)

    def test_marginally_bound(self):
        with self.assertRaises(ValueError): StableOrbit(0.5,12,1,0.5)

    def test_invalid_arguments(self):
        with self.assertRaises(ValueError): StableOrbit(2,5,0.5,0.5)
        with self.assertRaises(ValueError): StableOrbit(0.5,5,-0.5,0.5)

    def test_unstable(self):
        with self.assertRaises(ValueError): StableOrbit(0.5,5,0.5,0.5)

    def test_random(self):
        components = ["t","r","theta","phi"]
        orbit_values = np.genfromtxt(DATA_DIR / "orbit_values.txt", delimiter=",")
        orbit_times = np.genfromtxt(DATA_DIR / "orbit_times.txt", delimiter=",")

        for i, orbit in enumerate(orbit_values):
            mathematica_trajectory = np.genfromtxt(DATA_DIR / f"orbits/trajectory{i}.txt", delimiter=",")
            test_orbit = StableOrbit(*orbit)
            t, r , theta, phi = test_orbit.trajectory()
            python_trajectory = np.transpose(
                np.apply_along_axis(lambda x: np.array([t(x),r(x),theta(x),phi(x)]),0,orbit_times)
                )
            
            for j, component in enumerate(components):
                with self.subTest(i=i,
                                  component=component,
                                  params="a = {}, p = {}, e = {}, x = {}".format(*orbit),
                                  diff=np.max(np.abs(mathematica_trajectory[:,j]-python_trajectory[:,j]))
                                  ):
                    self.assertTrue(np.allclose(mathematica_trajectory[:,j],python_trajectory[:,j]))


"""
Module containing the Orbit class
"""
from .spacetime import KerrSpacetime
from .initial_conditions import *
from .units import *
from .constants import scale_constants
from .frequencies import mino_frequencies, fundamental_frequencies
from numpy import sin, cos, sqrt, pi
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.animation import FuncAnimation
from matplotlib.animation import FFMpegWriter

class Orbit:
    r"""
    Class representing an orbit in Kerr spacetime defined using initial conditions.

    :param a: spin parameter
    :type a: double
    :param initial_position: initial position of the orbit :math:`(t_0,r_0,\theta_0,\phi_0)`
    :type initial_position: tuple(double,double,double,double)
    :param initial_velocity: initial four-velocity of the orbit :math:`(u^t_0,u^r_0,u^\theta_0,u^\phi_0)`
    :type initial_velocity: tuple(double,double,double,double)
    :param M: mass of the primary in solar masses
    :type M: double, optional
    :param mu: mass of the smaller body in solar masses
    :type mu: double, optional

    :ivar a: spin parameter
    :ivar initial_position: initial position of the orbit :math:`(t_0,r_0,\theta_0,\phi_0)`
    :ivar initial_velocity: initial four-velocity of the orbit :math:`(u^t_0,u^r_0,u^\theta_0,u^\phi_0)`
    :ivar E: dimensionless energy
    :ivar L: dimensionless angular momentum
    :ivar Q: dimensionless carter constant
    :ivar stable: boolean indicating whether the orbit is stable
    :ivar upsilon_r: dimensionless radial orbital frequency in Mino time
    :ivar upsilon_theta: dimensionless polar orbital frequency in Mino time
    """
    def __init__(self,a,initial_position,initial_velocity, M=None, mu=None):
        self.a, self.initial_position, self.initial_velocity, self.M, self.mu = a, initial_position, initial_velocity, M, mu

        # check if initial four-velocity is valid
        spacetime = KerrSpacetime(a)
        initial_norm =  spacetime.norm(*initial_position,initial_velocity)
        if initial_norm >= 0: raise ValueError("Initial velocity is not timelike")
        if abs(initial_norm + 1) > 1e-6: raise ValueError("Initial velocity is not normalized")

        E, L, Q = constants_from_initial_conditions(a,initial_position,initial_velocity)
        self.E, self.L, self.Q = E, L, Q

        if is_stable(a,initial_position,initial_velocity):
            self.stable = True
            a, p, e, x = apex_from_constants(a,E,L,Q)
            self.a, self.p, self.e, self.x = a, p, e, x
            self.upsilon_r, self.upsilon_theta, self.upsilon_phi, self.gamma = mino_frequencies(a,p,e,x)
            self.omega_r, self.omega_theta, self.omega_phi = fundamental_frequencies(a,p,e,x)
            self.initial_phases = stable_orbit_initial_phases(a,initial_position,initial_velocity)
        else:
            self.stable = False
            self.upsilon_r, self.upsilon_theta = plunging_mino_frequencies(a,E,L,Q)
            self.initial_phases = plunging_orbit_initial_phases(a,initial_position,initial_velocity)
    
    def orbital_parameters(self):
        """
        Returns the orbital parameters :math:`(a,p,e,x)` of the orbit. Raises a ValueError if the orbit is not stable.

        :return: tuple of orbital parameters :math:`(a,p,e,x)`
        :rtype: tuple(double,double,double,double)
        """
        if not self.stable: raise ValueError("Orbit is not stable")
    
        return self.a, self.p, self.e, self.x

    def trajectory(self,initial_phases=None,distance_units="natural",time_units="natural"):
        r"""
        Computes the components of the trajectory as a function of Mino time
        :param initial_phases: tuple of initial phases :math:`(q_{t_0},q_{r_0},q_{\theta_0},q_{\phi_0})`
        :type initial_phases: tuple, optional
        :param distance_units: units to compute the radial component of the trajectory in (options are "natural", "mks", "cgs", "au" and "km"), defaults to "natural"
        :type distance_units: str, optional
        :param time_units: units to compute the time component of the trajectory in (options are "natural", "mks", "cgs", and "days"), defaults to "natural"
        :type time_units: str, optional

        :return: tuple of functions in the form :math:`(t(\lambda), r(\lambda), \theta(\lambda), \phi(\lambda))`
        :rtype: tuple(function, function, function, function)
        """
        if initial_phases is None: initial_phases = self.initial_phases
        if self.stable:
            from .stable_orbit import StableOrbit
            orbit = StableOrbit(self.a,self.p,self.e,self.x,M=self.M,mu=self.mu)
            return orbit.trajectory(initial_phases,distance_units,time_units)
        else:
            from .plunging_orbit import PlungingOrbit
            orbit = PlungingOrbit(self.a,self.E,self.L,self.Q,M=self.M,mu=self.mu)
            return orbit.trajectory(initial_phases,distance_units,time_units)
        
    def constants_of_motion(self, units="natural"):
        """
        Computes the energy, angular momentum, and carter constant for the orbit. Computes dimensionless constants in geometried units by default.
        M and mu must be defined in order to convert to physical units.

        :param units: units to return the constants of motion in (options are "natural", "mks" and "cgs"), defaults to "natural"
        :type units: str, optional

        :return: tuple of the form (E, L, Q)
        :rtype: tuple(double, double, double)
        """
        constants = self.E, self.L, self.Q
        if units == "natural":
            return constants
        
        if self.M is None or self.mu is None: raise ValueError("M and mu must be specified to convert constants of motion to physical units")
        
        if units == "mks":
            E, L, Q = scale_constants(constants,1,self.mu/self.M)
            return energy_in_joules(E,self.M), angular_momentum_in_mks(L,self.M), carter_constant_in_mks(Q,self.M)
        
        if units == "cgs":
            E, L, Q = scale_constants(constants,1,self.mu/self.M)
            return energy_in_ergs(E,self.M), angular_momentum_in_cgs(L,self.M), carter_constant_in_cgs(Q,self.M)
        
        raise ValueError("units must be one of 'natural', 'mks', or 'cgs'")

    def four_velocity(self,initial_phases=None):
        r"""
        Computes the four velocity of the orbit as a function of Mino time

        :param initial_phases: tuple of initial phases :math:`(q_{t_0},q_{r_0},q_{\theta_0},q_{\phi_0})`
        :type initial_phases: tuple, optional

        :return: components of the four velocity (i.e. :math:`u^t,u^r,u^\theta,u^\phi`)
        :rtype: tuple(function, function, function, function)
        """
        if initial_phases is None: initial_phases = self.initial_phases
        t, r, theta, phi = self.trajectory(initial_phases)
        spacetime = KerrSpacetime(self.a)
        constants = self.E, self.L, self.Q

        return spacetime.four_velocity(t,r,theta,phi,constants,self.upsilon_r,self.upsilon_theta,initial_phases)
    
    def four_velocity_norm(self,initial_phases=None):
        r"""
        Computes the norm of the four velocity of the orbit as a function of Mino time

        :param initial_phases: tuple of initial phases :math:`(q_{t_0},q_{r_0},q_{\theta_0},q_{\phi_0})`
        :type initial_phases: tuple, optional

        :return: norm of the four velocity :math:`g_{\mu\nu}u^\mu u^\nu`
        :rtype: function
        """
        if initial_phases is None: initial_phases = self.initial_phases
        t, r, theta, phi = self.trajectory(initial_phases)
        spacetime = KerrSpacetime(self.a)
        constants = self.E, self.L, self.Q
        t_prime, r_prime, theta_prime, phi_prime = self.four_velocity(initial_phases)

        def norm(time):
            u = [t_prime(time),r_prime(time),theta_prime(time),phi_prime(time)]
            return spacetime.norm(t(time),r(time),theta(time),phi(time),u)

        return norm

    def plot(self,lambda0=0, lambda1=20, elevation=30 ,azimuth=-60, initial_phases=None, grid=True, axes=True, thickness=1):
        r"""
        Creates a plot of the orbit

        :param lambda0: starting mino time
        :type lambda0: double, optional
        :param lambda1: ending mino time
        :type lambda1: double, optional
        :param elevation: camera elevation angle in degrees
        :type elevation: double, optional
        :param azimuth: camera azimuthal angle in degrees
        :type azimuth: double, optional
        :param initial_phases: tuple of initial phases :math:`(q_{t_0},q_{r_0},q_{\theta_0},q_{\phi_0})`
        :type initial_phases: tuple, optional
        :param grid: if true, grid lines are shown on plot
        :type grid: bool, optional
        :param axes: if true, axes are shown on plot
        :type axes: bool, optional
        :param thickness: line thickness of the orbit
        :type thickness: double, optional

        :return: matplotlib figure and axes
        :rtype: matplotlib.figure.Figure, matplotlib.axes._subplots.AxesSubplot
        """
        if initial_phases is None: initial_phases = self.initial_phases
        lambda_range = lambda1 - lambda0
        point_density = 500
        num_pts = int(lambda_range*point_density)
        time = np.linspace(lambda0,lambda1,num_pts)

        t, r, theta, phi = self.trajectory(initial_phases=initial_phases)

        # compute trajectory in cartesian coordinates
        trajectory_x = r(time)*sin(theta(time))*cos(phi(time))
        trajectory_y = r(time)*sin(theta(time))*sin(phi(time))
        trajectory_z = r(time)*cos(theta(time))
        trajectory = np.column_stack((trajectory_x,trajectory_y,trajectory_z))

        # create sphere with radius equal to event horizon radius
        event_horizon = 1+sqrt(1-self.a**2)
        u = np.linspace(0, 2 * np.pi, 100)
        v = np.linspace(0, np.pi, 100)
        x_sphere = event_horizon * np.outer(np.cos(u), np.sin(v))
        y_sphere = event_horizon * np.outer(np.sin(u), np.sin(v))
        z_sphere = event_horizon * np.outer(np.ones(np.size(u)), np.cos(v))

        # convert viewing angles to radians
        elevation_rad = elevation*pi/180
        azimuth_rad = azimuth*pi/180

        # https://matplotlib.org/stable/api/toolkits/mplot3d/view_angles.html
        view_plane_normal = [cos(elevation_rad)*cos(azimuth_rad),cos(elevation_rad)*sin(azimuth_rad),sin(elevation_rad)]
        # matplotlib has no ray tracer so points behind the black hole must be filtered out manually
        # for each trajectory point compute the component normal to the viewing plane
        normal_component = np.apply_along_axis(lambda x: np.dot(view_plane_normal,x),1,trajectory)
        # compute the projection of each trajectory point onto the viewing plane
        projection = trajectory-np.transpose(normal_component*np.transpose(np.broadcast_to(view_plane_normal,(num_pts,3))))
        # find points in front of the viewing plane or outside the event horizon when projected onto the viewing plane
        condition = (np.linalg.norm(trajectory,axis=1) > event_horizon) & ((normal_component >= 0) | (np.linalg.norm(projection,axis=1) > event_horizon))
        x_visible = trajectory_x[condition]
        y_visible = trajectory_y[condition]
        z_visible = trajectory_z[condition]

        fig = plt.figure()
        ax = fig.add_subplot(projection='3d')
        
        # plot black hole
        ax.plot_surface(x_sphere, y_sphere, z_sphere, color='black',shade=False)
        # plot orbit
        ax.scatter(x_visible,y_visible,z_visible,color="red",s=thickness)

        # set viewing angle
        ax.view_init(elevation,azimuth)
        # set equal aspect ratio and orthogonal projection
        ax.set_box_aspect([np.ptp(x_visible),np.ptp(y_visible),np.ptp(z_visible)])
        # https://matplotlib.org/stable/gallery/mplot3d/projections.html
        ax.set_proj_type('ortho')

        # turn off grid and axes if specified
        if not grid: ax.grid(False)
        if not axes: ax.axis("off")

        return fig, ax
    
    def is_visible(self,point,elevation,azimuth):
        """
        Determines if a point is visible from a given viewing angle or obscured by the black hole. 
        Viewing angles are defined as in https://matplotlib.org/stable/api/toolkits/mplot3d/view_angles.html and black hole is centered at the origin.

        :param point: cartersian coordinates of point to test
        :type point: array_like
        :param elevation: camera elevation angle in degrees
        :type elevation: double
        :param azimuth: camera azimuthal angle in degrees
        :type azimuth: double
        """
        # compute event horizon radius
        event_horizon = 1+sqrt(1-self.a**2)

        # convert viewing angles to radians
        elevation_rad = elevation*pi/180
        azimuth_rad = azimuth*pi/180

        # https://matplotlib.org/stable/api/toolkits/mplot3d/view_angles.html
        view_plane_normal = [cos(elevation_rad)*cos(azimuth_rad),cos(elevation_rad)*sin(azimuth_rad),sin(elevation_rad)]

        # compute the component normal to the viewing plane
        normal_component = np.dot(view_plane_normal,point)
        # compute the projection of the point onto the viewing plane
        projection = point-normal_component*view_plane_normal

        # test if point is outside the event horizon and either in front of the viewing plane or outside the event horizon when projected onto the viewing plane
        return True if (np.linalg.norm(point) > event_horizon) & ((normal_component >= 0) | (np.linalg.norm(projection) > event_horizon)) else False
    
    def animate(self,filename,lambda0=0, lambda1=10, elevation=30 ,azimuth=-60, initial_phases=None, grid=True, axes=True, thickness=2, tail_length="long"):
        r"""
        Saves an animation of the orbit as an mp4 file. 
        Note that this function requires ffmpeg to be installed and may take several minutes to run depending on the length of the animation.

        :param filename: filename to save the animation to
        :type filename: str
        :param lambda0: starting mino time, defaults to 0
        :type lambda0: double, optional
        :param lambda1: ending mino time, defaults to 10
        :type lambda1: double, optional
        :param elevation: camera elevation angle in degrees, defaults to 30
        :type elevation: double, optional
        :param azimuth: camera azimuthal angle in degrees, defaults to -60
        :type azimuth: double, optional
        :param initial_phases: tuple of initial phases :math:`(q_{t_0},q_{r_0},q_{\theta_0},q_{\phi_0})`
        :type initial_phases: tuple, optional
        :param grid: sets visibility of the grid, defaults to True
        :type grid: bool, optional
        :param axes: sets visibility of axes, defaults to True
        :type axes: bool, optional
        :param thickness: thickness of the tail of the orbit, defaults to 2
        :type thickness: double, optional
        :param tail: sets the length of the tail (options are "short", "long" and "none"), defaults to "long"
        :type tail: str, optional
        """
        lambda_range = lambda1 - lambda0
        point_density = 200
        num_pts = int(lambda_range*point_density)
        time = np.linspace(lambda0,lambda1,num_pts)

        fig = plt.figure(figsize=(7,7))
        ax = fig.add_subplot(projection='3d')
        eh = 1+sqrt(1-self.a**2)

        body = ax.scatter([],[],[],c="black")
        tail = ax.scatter([],[],[],c="red", s=thickness)

        t, r, theta, phi = self.trajectory(initial_phases)

        ax.view_init(elevation,azimuth)
        x = r(time)*sin(theta(time))*cos(phi(time))
        y = r(time)*sin(theta(time))*sin(phi(time))
        z = r(time)*cos(theta(time))
        trajectory = np.column_stack((x,y,z))

        azimuth_rad = azimuth*pi/180
        elevation_rad = elevation*pi/180

        # https://matplotlib.org/stable/api/toolkits/mplot3d/view_angles.html
        view_plane_normal = [cos(elevation_rad)*cos(azimuth_rad),cos(elevation_rad)*sin(azimuth_rad),sin(elevation_rad)]

        # matplotlib has no ray tracer so points behind the black hole must be filtered out manually
        # for each trajectory point compute the component normal to the viewing plane
        normal_component = np.apply_along_axis(lambda x: np.dot(view_plane_normal,x),1,trajectory)
        # compute the projection of each trajectory point onto the viewing plane
        projection = trajectory-np.transpose(normal_component*np.transpose(np.broadcast_to(view_plane_normal,(num_pts,3))))
        # find points in front of the viewing plane or outside the event horizon when projected onto the viewing plane
        condition = (np.dot(trajectory,view_plane_normal) >= 0) | (np.linalg.norm(projection,axis=1) > eh)

        # create sphere with radius equal to event horizon radius
        u = np.linspace(0, 2 * np.pi, 100)
        v = np.linspace(0, np.pi, 100)
        x_sphere = eh * np.outer(np.cos(u), np.sin(v))
        y_sphere = eh * np.outer(np.sin(u), np.sin(v))
        z_sphere = eh * np.outer(np.ones(np.size(u)), np.cos(v))

        # plot black hole
        ax.plot_surface(x_sphere, y_sphere, z_sphere, color='black',shade=False,zorder=0)

        # set axis limits
        ax.set_xlim([x.min(),x.max()])
        ax.set_ylim([y.min(),y.max()])
        ax.set_zlim([z.min(),z.max()])

        # set equal aspect ratio and orthogonal projection
        ax.set_box_aspect([np.ptp(x),np.ptp(y),np.ptp(z)])
        # https://matplotlib.org/stable/gallery/mplot3d/projections.html
        ax.set_proj_type('ortho')

         # turn off grid and axes if specified
        if not grid: ax.grid(False)
        if not axes: ax.axis("off")

        def animate(i,body,tail):
            # adjust length of tail
            start = 0
            if tail_length == "short": start = max(0,i-50)
            elif tail_length == "none": start = i

            condition_slice = condition[start:i]
            body._offsets3d = ([x[i]],[y[i]],[z[i]])
            tail._offsets3d = (x[start:i][condition_slice],y[start:i][condition_slice],z[start:i][condition_slice])
            
        # save to file
        ani = FuncAnimation(fig,animate,num_pts,fargs=(body,tail))
        FFwriter = FFMpegWriter(fps=60)
        ani.save(filename, writer = FFwriter)


"""
Module containing functions for computing the constants of motion for bound orbits using the :math:`(a,p,e,x)` parametrization.
Constants of motion are computed using the method described in Appendix B of `Schmidt <https://doi.org/10.48550/arXiv.gr-qc/0202090>`_.
"""
from numpy import sign, sqrt, copysign, pi, nan, inf
from scipy.optimize import root_scalar
from .units import *
from scipy.interpolate import RectBivariateSpline
import numpy as np

def _coefficients(r,a,x):
    """
    Computes the coefficients f, g, h and d from equation B.5 in `Schmidt <https://doi.org/10.48550/arXiv.gr-qc/0202090>`_
    
    :param r: dimensionless distance from the black hole
    :type r: double
    :param a: dimensionless spin of the black hole
    :type a: double
    :param x: cosine of the orbital inclination
    :type x: double
    
    :rtype: tuple(double, double, double, double)
    """
    z = sqrt(1-x**2)
    delta = r**2-2*r+a**2
    f = lambda r: r**4+a**2*(r*(r+2)+z**2*delta)
    g = lambda r: 2*a*r
    h = lambda r: r*(r-2)+z**2/(1-z**2)*delta
    d = lambda r: (r**2+a**2*z**2)*delta
    
    return f(r), g(r), h(r), d(r)

def _coefficients_derivative(r,a,x):
    """
    Computes the derivatives f', g', h' and d' of the coefficients from equation B.5 in `Schmidt <https://doi.org/10.48550/arXiv.gr-qc/0202090>`_
    
    :param r: dimensionless distance from the black hole
    :type r: double
    :param a: dimensionless spin of the black hole
    :type a: double
    :param x: cosine of the orbital inclination
    :type x: double
    
    :rtype: tuple(double, double, double, double)
    """
    z = sqrt(1-x**2)
    f_prime = lambda r: 4*r**3+2*a**2*((1+z**2)*r+(1-z**2))
    g_prime = lambda r: 2*a
    h_prime = lambda r: 2*(r-1)/(1-z**2)
    d_prime = lambda r: 2*(2*r-3)*r**2+2*a**2*((1+z**2)*r-z**2)
    
    return f_prime(r), g_prime(r), h_prime(r), d_prime(r)

def _standardize_params(a,x):
    """
    Changes signs of a and x so that a is positive and x encodes the direction of the orbit.

    :param a: dimensionless spin of the black hole
    :type a: double
    :param x: cosine of the orbital inclination
    :type x: double

    :rtype: tuple(double, double)
    """
    return abs(a), x*copysign(1,a)

def energy(a,p,e,x):
    """
    Computes the dimensionless energy of a bound orbit with the given parameters
    
    :param a: dimensionless spin of the black hole (must satisfy -1 < a < 1)
    :type a: double
    :param p: orbital semi-latus rectum
    :type p: double
    :param e: orbital eccentricity (must satisfy 0 <= e <= 1)
    :type e: double
    :param x: cosine of the orbital inclination (must satisfy 0 <= x^2 <= 1)
    :type x: double
    
    :rtype: double
    """
    a, x = _standardize_params(a,x)

    if a == 1: raise ValueError("Extreme Kerr not supported")
    if not valid_params(a,e,x): raise ValueError("a^2, e and x^2 must be between 0 and 1")
    if not is_stable(a,p,e,x): raise ValueError("Not a stable orbit")

    # marginally bound case
    if e == 1:
        return 1
    
    # polar case
    if x == 0:
        # expression from ConstantsOfMotion.m in the KerrGeodesics mathematica library
        return sqrt(-((p*(a**4*(-1 + e**2)**2 + (-4*e**2 + (-2 + p)**2)*p**2 + \
                2*a**2*p*(-2 + p + e**2*(2 + p))))/ \
                (a**4*(-1 + e**2)**2*(-1 + e**2 - p) + (3 + e**2 - p)*p**4 - \
                2*a**2*p**2*(-1 - e**4 + p + e**2*(2 + p)))))
    
    # spherical case
    if e == 0:
        r0 = p
        f1, g1, h1, d1 = _coefficients(r0,a,x)
        f2, g2, h2, d2 = _coefficients_derivative(r0,a,x)
    # generic case
    else:
        r1 = p/(1-e)
        r2 = p/(1+e)
        f1, g1, h1, d1 = _coefficients(r1,a,x)
        f2, g2, h2, d2 = _coefficients(r2,a,x)
    
    # equation B.19 - B.21
    kappa = d1*h2-h1*d2
    rho = f1*h2-h1*f2
    sigma = g1*h2-h1*g2
    epsilon = d1*g2-g1*d2
    eta = f1*g2-g1*f2
    
    # equation B.22
    return sqrt(
                (kappa*rho+2*epsilon*sigma-sign(x)*2*sqrt(sigma*(sigma*epsilon**2+rho*epsilon*kappa-eta*kappa**2)))
                /(rho**2+4*eta*sigma)
               )

def angular_momentum(a,p,e,x,E=None):
    """
    Computes the dimensionless angular momentum of a bound orbit with the given parameters
    
    :param a: dimensionless spin of the black hole (must satisfy -1 < a < 1)
    :type a: double
    :param p: orbital semi-latus rectum
    :type p: double
    :param e: orbital eccentricity (must satisfy 0 <= e <= 1)
    :type e: double
    :param x: cosine of the orbital inclination (must satisfy 0 <= x^2 <= 1)
    :type x: double
    :param E: dimensionless energy of the orbit can be passed in to speed computation if it is already known
    :type E: double, optional
    
    :rtype: double
    """
    a, x = _standardize_params(a,x)

    if a == 1: raise ValueError("Extreme Kerr not supported")
    if not valid_params(a,e,x): raise ValueError("a^2, e and x^2 must be between 0 and 1")
    if not is_stable(a,p,e,x): raise ValueError("Not a stable orbit")

    # compute energy if not given
    if E is None: E = energy(a,p,e,x)

    # polar case
    if x == 0:
        return 0
    
    # marginally bound case
    if e == 1:
        r2 = p/(1+e)
        f2, g2, h2, d2 = _coefficients(r2,a,x)
        # obtained by solving equation B.17 for L
        return (-E*g2 + sign(x)*sqrt(-d2*h2 + E**2*(g2**2 + f2*h2)))/h2
    
    # generic case
    else:
        r1 = p/(1-e)
        f1, g1, h1, d1 = _coefficients(r1,a,x)
        # obtained by solving equation B.17 for L
        return (-E*g1 + sign(x)*sqrt(-d1*h1 + E**2*(g1**2 + f1*h1)))/h1
    
def carter_constant(a,p,e,x,E=None,L=None):
    """
    Computes the dimensionless carter constant of a bound orbit with the given parameters
    
    :param a: dimensionless spin of the black hole (must satisfy -1 < a < 1)
    :type a: double
    :param p: orbital semi-latus rectum
    :type p: double
    :param e: orbital eccentricity (must satisfy 0 <= e <= 1)
    :type e: double
    :param x: cosine of the orbital inclination (must satisfy 0 <= x^2 <= 1)
    :type x: double
    :param E: dimensionless energy of the orbit can be passed in to speed computation if it is already known
    :type E: double, optional
    :param L: dimensionless angular momentum of the orbit can be passed in to speed computation if it is already known
    :type L: double, optional
    
    :rtype: double
    """
    a, x = _standardize_params(a,x)

    if a == 1:  raise ValueError("Extreme Kerr not supported")
    if not valid_params(a,e,x): raise ValueError("a^2, e and x^2 must be between 0 and 1")
    if not is_stable(a,p,e,x): raise ValueError("Not a stable orbit")

    # polar case
    if x == 0:
        # expression from ConstantsOfMotion.m in the KerrGeodesics mathematica library
        return -((p**2*(a**4*(-1+e**2)**2+p**4+2*a**2*p*(-2+p+e**2*(2+p)))) \
                 /(a**4*(-1+e**2)**2*(-1+e**2-p)+(3+e**2-p)*p**4-2*a**2*p**2*(-1-e**4+p+e**2*(2+p))))
    
    z = sqrt(1-x**2)
    # compute energy and angular momentum if not given
    if E is None: E = energy(a,p,e,x)
    if L is None: L = angular_momentum(a,p,e,x,E)
    #  equation B.4
    return z**2 * (a**2 * (1 - E**2) + L**2/(1 - z**2))

def constants_of_motion(a,p,e,x):
    """
    Computes the dimensionless energy, angular momentum, and Carter constant of a bound orbit with the given parameters
    
    :param a: dimensionless spin of the black hole (must satisfy -1 < a < 1)
    :type a: double
    :param p: orbital semi-latus rectum
    :type p: double
    :param e: orbital eccentricity (must satisfy 0 <= e <= 1)
    :type e: double
    :param x: cosine of the orbital inclination (must satisfy 0 <= x^2 <= 1)
    :type x: double
    
    :return: tuple of constants in the form :math:`(E, L, Q)`
    :rtype: tuple(double, double, double)
    """
    E = energy(a,p,e,x)
    L = angular_momentum(a,p,e,x,E)
    Q = carter_constant(a,p,e,x,E,L)
    return E, L, Q

def _S_polar(p,a,e):
    """
    Separatrix polynomial for a polar orbit from equation 37 in `Stein and Warburton <https://doi.org/10.48550/arXiv.1912.07609>`_

    :param p: orbital semi-latus rectum
    :type p: double
    :param a: dimensionless spin of the black hole
    :type a: double
    :param e: orbital eccentricity
    :type e: double

    :rtype: double
    """
    return      p**5*(-6 - 2*e + p) \
                + a**2*p**3*(-4*(-1+e)*(1+e)**2 + (3 + e*(2 + 3*e))*p) \
                - a**4*(1 + e)**2*p*(6 + 2*e**3 + 2*e*(-1 + p) - 3*p - 3*e**2*(2 + p)) \
                + a**6*(-1 + e)**2*(1 + e)**4

def _S_equatorial(p,a,e):
    """
    Separatrix polynomial for an equatorial orbit from equation 23 in `Stein and Warburton <https://doi.org/10.48550/arXiv.1912.07609>`_

    :param p: orbital semi-latus rectum
    :type p: double
    :param a: dimensionless spin of the black hole
    :type a: double
    :param e: orbital eccentricity
    :type e: double

    :rtype: double
    """
    return      a**4*(-3 - 2*e + e**2)**2 \
                + p**2*(-6 - 2*e + p)**2 \
                - 2*a**2*(1 + e)*p*(14 + 2*e**2 + 3*p - e*p)

def _S(p,a,e,x):
    """
    Full separatrix polynomial from equation A1 in `Stein and Warburton <https://doi.org/10.48550/arXiv.1912.07609>`_

    :param p: orbital semi-latus rectum
    :type p: double
    :param a: dimensionless spin of the black hole
    :type a: double
    :param e: orbital eccentricity
    :type e: double

    :rtype: double
    """
    return -4*(3 + e)*p**11 + p**12 + \
       a**12*(-1 + e)**4*(1 + e)**8*(-1 + x)**4*(1 + x)**4 - \
       4*a**10*(-3 + e)*(-1 + e)**3*(1 + e)**7*p*(-1 + x**2)**4 - \
       4*a**8*(-1 + e)*(1 + e)**5*p**3*(-1 + x)**3*(1 + x)**3* \
        (7 - 7*x**2 - e**2*(-13 + x**2) + e**3*(-5 + x**2) + 7*e*(-1 + x**2)) + \
       8*a**6*(-1 + e)*(1 + e)**3*p**5*(-1 + x**2)**2* \
        (3 + e + 12*x**2 + 4*e*x**2 + e**3*(-5 + 2*x**2) + e**2*(1 + 2*x**2)) - \
       8*a**4*(1 + e)**2*p**7*(-1 + x)*(1 + x)* \
        (-3 + e + 15*x**2 - 5*e*x**2 + e**3*(-5 + 3*x**2) + e**2*(-1 + 3*x**2))\
        + 4*a**2*p**9*(-7 - 7*e + e**3*(-5 + 4*x**2) + e**2*(-13 + 12*x**2)) + \
       2*a**8*(-1 + e)**2*(1 + e)**6*p**2*(-1 + x**2)**3* \
        (2*(-3 + e)**2*(-1 + x**2) + \
          a**2*(e**2*(-3 + x**2) - 3*(1 + x**2) + 2*e*(1 + x**2))) - \
       2*p**10*(-2*(3 + e)**2 + a**2* \
           (-3 + 6*x**2 + e**2*(-3 + 2*x**2) + e*(-2 + 4*x**2))) + \
       a**6*(1 + e)**4*p**4*(-1 + x**2)**2* \
        (-16*(-1 + e)**2*(-3 - 2*e + e**2)*(-1 + x**2) + \
          a**2*(15 + 6*x**2 + 9*x**4 + e**2*(26 + 20*x**2 - 2*x**4) + \
             e**4*(15 - 10*x**2 + x**4) + 4*e**3*(-5 - 2*x**2 + x**4) - \
             4*e*(5 + 2*x**2 + 3*x**4))) - \
       4*a**4*(1 + e)**2*p**6*(-1 + x)*(1 + x)* \
        (-2*(11 - 14*e**2 + 3*e**4)*(-1 + x**2) + \
          a**2*(5 - 5*x**2 - 9*x**4 + 4*e**3*x**2*(-2 + x**2) + \
             e**4*(5 - 5*x**2 + x**4) + e**2*(6 - 6*x**2 + 4*x**4))) + \
       a**2*p**8*(-16*(1 + e)**2*(-3 + 2*e + e**2)*(-1 + x**2) + \
          a**2*(15 - 36*x**2 + 30*x**4 + e**4*(15 - 20*x**2 + 6*x**4) + \
             4*e**3*(5 - 12*x**2 + 6*x**4) + 4*e*(5 - 12*x**2 + 10*x**4) + \
             e**2*(26 - 72*x**2 + 44*x**4)))

def separatrix(a,e,x):
    """
    Returns the value of p at the separatrix for the given orbital parameters computed using the bracked root finding method described in `Stein and Warburton <https://doi.org/10.48550/arXiv.1912.07609>`_
    
    :param a: dimensionless spin of the black hole (must satisfy 0 <= a < 1)
    :type a: double
    :param e: orbital eccentricity (must satisfy 0 <= e <= 1)
    :type e: double
    :param x: cosine of the orbital inclination (must satisfy 0 <= x^2 <= 1)
    :type x: double
    
    :rtype: double
    """

    if a == 1: raise ValueError("Extreme Kerr not supported")
    if not valid_params(a,e,x): raise ValueError("a^2, e and x^2 must be between 0 and 1")

    if a == 0:
        return 6+2*e
    
    polar_bracket = [1+sqrt(3)+sqrt(3+2*sqrt(3)), 8]
    p_polar = root_scalar(_S_polar, args=(a,e), bracket=polar_bracket)
    
    if x == 0:
        return p_polar.root
        
    equatorial_prograde_bracket = [1+e, 6+2*e]
    p_equatorial_prograde = root_scalar(_S_equatorial,args=(a,e),bracket=equatorial_prograde_bracket)
    
    if x == 1: 
        return p_equatorial_prograde.root
    
    if x == -1:
        equatorial_retrograde_bracket = [6+2*e, 5+e+4*sqrt(1+e)]
        p_equatorial_retrograde = root_scalar(_S_equatorial,args=(a,e),bracket=equatorial_retrograde_bracket)
        return p_equatorial_retrograde.root
    
    if x > 0:
        p = root_scalar(_S,args=(a,e,x),bracket=[p_equatorial_prograde.root, p_polar.root])
        return p.root
    
    if x < 0:
        p = root_scalar(_S,args=(a,e,x),bracket=[p_polar.root, 12])
        return p.root
    
def fast_separatrix(a, grid_spacing=0.01):
    """
    Constructs a faster separatrix function for a given value of :math:`a` by interpolating over a grid of :math:`e` and :math:`x` values.

    :param a: dimensionless spin of the black hole
    :type a: double
    :param grid_spacing: spacing of the grid over which to interpolate, defaults to 0.01
    :type grid_spacing: double, optional

    :return: interpolated function of e and x
    :rtype: scipy.interpolate.RectBivariateSpline
    """
    
    # create grid of e and x values to interpolate over
    num_e_pts = int(1/grid_spacing)
    num_x_pts = int(2/grid_spacing)
    e = np.linspace(0,1,num_e_pts)
    x = np.linspace(-1,1,num_x_pts)
    E, X = np.meshgrid(e,x)

    # compute separatrix values on grid
    P = np.zeros((num_e_pts,num_x_pts))
    for i in range(num_e_pts):
        for j in range(num_x_pts):
            P[i,j] = separatrix(a,E[j,i],X[j,i])

    # create interpolator
    interpolator = RectBivariateSpline(e,x,P)

    return interpolator

def is_stable(a,p,e,x):
    """
    Tests whether or not the given orbital parameters define a stable bound orbit
    
    :param a: dimensionless spin of the black hole
    :type a: double
    :param p: orbital semi-latus rectum
    :type p: double
    :param e: orbital eccentricity
    :type e: double
    :param x: cosine of the orbital inclination
    :type x: double
    
    :rtype: boolean
    """
    if p > separatrix(a,e,x):
        return True
    return False

def valid_params(a,e,x):
    """
    Tests whether the given parameters fall into the allowed ranges

    :param a: dimensionless spin of the black hole
    :type a: double
    :param e: orbital eccentricity
    :type e: double
    :param x: cosine of the orbital inclination
    :type x: double

    :rtype: boolean
    """
    if (0 <= a <= 1) and (0 <= e <= 1) and (-1 <= x <= 1):
        return True
    return False

def scale_constants(constants,M,mu):
    """
    Scales the dimensionless constants of motion to the given mass parameters
    
    :param constants: dimensionless constants of motion in the form (E, L, Q)
    :type constants: tuple
    :param M: mass of the black hole
    :type M: double
    :param mu: mass ratio
    :type mu: double
    
    :rtype: tuple(double, double, double)
    """
    M = mass_in_kg(M)
    return constants[0]*mu, constants[1]*mu*M, constants[2]*mu**2*M**2


"""
Python package for computing plunging and non-plunging geodesics in Kerr spacetime.
"""
__all__ = ["units","constants", "frequencies", "stable_orbit", "plunging_orbit","initial_conditions"]
from kerrgeopy import *
from kerrgeopy.frequencies import *
from kerrgeopy.constants import *
from kerrgeopy.initial_conditions import *
from kerrgeopy.stable_orbit import StableOrbit
from kerrgeopy.plunging_orbit import PlungingOrbit
from kerrgeopy.orbit import Orbit
from kerrgeopy.spacetime import KerrSpacetime


from setuptools import setup
import setuptools

setup(
    name="kerrgeopy",
    version="0.2.0",
    author="Seyong Park",
    description="Library for computing bound and plunging geodesics in Kerr spacetime",
    url="https://github.com/syp2001/KerrGeoPy",
    packages=setuptools.find_packages(exclude=["tests"]),
    classifiers=(
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License (GPL)",
        "Natural Language :: English",
    ),
    install_requires=["scipy>=1.8","numpy","matplotlib>=3.3"]
)


"""
Module containing the BoundOrbit class
"""
from .constants import *
from .frequencies import *
from .stable_solutions import *
from .initial_conditions import apex_from_constants
from .frequencies import _radial_roots, _polar_roots
from .orbit import Orbit

class StableOrbit(Orbit):
    r"""
    Class representing a stable bound orbit in Kerr spacetime.

    :param a: dimensionless angular momentum (must satisfy 0 <= a < 1)
    :type a: double
    :param p: semi-latus rectum
    :type p: double
    :param e: orbital eccentricity (must satisfy 0 <= e < 1)
    :type e: double
    :param x: cosine of the orbital inclination (must satisfy 0 < x^2 <= 1)
    :type x: double
    :param M: mass of the primary in solar masses, optional
    :type M: double
    :param mu: mass of the smaller body in solar masses, optional
    :type mu: double

    :ivar a: dimensionless angular momentum
    :ivar p: semi-latus rectum
    :ivar e: orbital eccentricity
    :ivar x: cosine of the orbital inclination
    :ivar M: mass of the primary in solar masses
    :ivar mu: mass of the smaller body in solar masses
    :ivar E: dimensionless energy
    :ivar L: dimensionless angular momentum
    :ivar Q: dimensionless carter constant
    :ivar initial_position: tuple of initial position coordinates :math:`(t_0, r_0, \theta_0, \phi_0)`
    :ivar initial_velocity: tuple of initial four-velocity components :math:`(u^t_0, u^r_0, u^\theta_0, u^\phi_0)`
    :ivar upsilon_r: dimensionless radial orbital frequency in Mino time
    :ivar upsilon_theta: dimensionless polar orbital frequency in Mino time
    :ivar upsilon_phi: dimensionless azimuthal orbital frequency in Mino time
    :ivar gamma: dimensionless time dilation factor
    :ivar omega_r: dimensionless radial orbital frequency in Boyer-Lindquist time
    :ivar omega_theta: dimensionless polar orbital frequency in Boyer-Lindquist time
    :ivar omega_phi: dimensionless azimuthal orbital frequency in Boyer-Lindquist time
    """
    def __init__(self,a,p,e,x,initial_phases=(0,0,0,0),M = None,mu=None):
        """
        Constructor method
        """
        self.a, self.p, self.e, self.x, self.initial_phases, self.M, self.mu = a, p, e, x, initial_phases, M, mu
        constants = constants_of_motion(a,p,e,x)

        self.E, self.L, self.Q = constants
        self.upsilon_r, self.upsilon_theta, self.upsilon_phi, self.gamma = mino_frequencies(a,p,e,x)
        self.omega_r, self.omega_theta, self.omega_phi = fundamental_frequencies(a,p,e,x)
        self.stable = True

        u_t, u_r, u_theta, u_phi = self.four_velocity()
        t, r, theta, phi = self.trajectory()
        self.initial_position = t(0), r(0), theta(0), phi(0)
        self.initial_velocity = u_t(0), u_r(0), u_theta(0), u_phi(0)

    @classmethod
    def from_constants(cls,a,E,L,Q,initial_phases=(0,0,0,0),M=None,mu=None):
        """
        Alternative constructor method that takes the constants of motion as arguments.

        :param a: spin parameter
        :type a: double
        :param E: dimensionless energy
        :type E: double
        :param L: dimensionless angular momentum
        :type L: double
        :param Q: dimensionless Carter constant
        :type Q: double
        :param M: mass of the primary in solar masses
        :type M: double, optional
        :param mu: mass of the smaller body in solar masses
        :type mu: double, optional
        """
        a, p, e, x = apex_from_constants(a,E,L,Q)

        return cls(a,p,e,x,initial_phases,M,mu)
        
    def mino_frequencies(self, units="natural"):
        r"""
        Computes orbital frequencies in Mino time. Returns dimensionless frequencies in geometrized units by default.
        M and mu must be defined in order to convert to physical units.

        :param units: units to return the frequencies in (options are "natural", "mks" and "cgs"), defaults to "natural"
        :type units: str, optional

        :return: tuple of orbital frequencies in the form :math:`(\Upsilon_r, \Upsilon_\theta, \Upsilon_\phi, \Gamma)`
        :rtype: tuple(double, double, double, double)
        """
        upsilon_r, upsilon_theta, upsilon_phi, gamma = self.upsilon_r, self.upsilon_theta, self.upsilon_phi, self.gamma
        if units == "natural":
            return upsilon_r, upsilon_theta, upsilon_phi, gamma
        
        if self.M is None: raise ValueError("M must be specified to convert frequencies to physical units")
        
        if units == "mks" or units == "cgs":
            return time_in_seconds(upsilon_r,self.M), time_in_seconds(upsilon_theta,self.M), time_in_seconds(upsilon_phi,self.M), time2_in_seconds2(gamma,self.M)
        
        raise ValueError("units must be one of 'natural', 'mks', or 'cgs'")
    
    def fundamental_frequencies(self, units="natural"):
        r"""
        Computes orbital frequencies in Boyer-Lindquist time. Returns dimensionless frequencies in geometrized units by default.
        M and mu must be defined in order to convert to physical units.

        :param units: units to return the frequencies in (options are "natural", "mks", "cgs" and "mHz"), defaults to "natural"
        :type units: str, optional
        :return: tuple of orbital frequencies in the form :math:`(\Omega_r, \Omega_\theta, \Omega_\phi)`
        :rtype: tuple(double, double, double)
        """
        upsilon_r, upsilon_theta, upsilon_phi, gamma = self.upsilon_r, self.upsilon_theta, self.upsilon_phi, self.gamma
        if units == "natural":
            return upsilon_r/gamma, upsilon_theta/gamma, upsilon_phi/gamma
        
        if self.M is None: raise ValueError("M must be specified to convert frequencies to physical units")

        if units == "mks" or units == "cgs":
            return frequency_in_Hz(upsilon_r/gamma,self.M), frequency_in_Hz(upsilon_theta/gamma,self.M), frequency_in_Hz(upsilon_phi/gamma,self.M)
        if units == "mHz":
            return frequency_in_mHz(upsilon_r/gamma,self.M), frequency_in_mHz(upsilon_theta/gamma,self.M), frequency_in_mHz(upsilon_phi/gamma,self.M)
        
        raise ValueError("units must be one of 'natural', 'mks', 'cgs', or 'mHz'")
        
    def trajectory(self,initial_phases=None,distance_units="natural",time_units="natural"):
        r"""
        Computes the time, radial, polar, and azimuthal coordinates of the orbit as a function of mino time.

        :param initial_phases: tuple of initial phases :math:`(q_{t_0},q_{r_0},q_{\theta_0},q_{\phi_0})`, defaults to (0,0,0,0)
        :type initial_phases: tuple, optional
        :param distance_units: units to compute the radial component of the trajectory in (options are "natural", "mks", "cgs", "au" and "km"), defaults to "natural"
        :type distance_units: str, optional
        :param time_units: units to compute the time component of the trajectory in (options are "natural", "mks", "cgs", and "days"), defaults to "natural"
        :type time_units: str, optional

        :return: tuple of functions in the form :math:`(t(\lambda), r(\lambda), \theta(\lambda), \phi(\lambda))`
        :rtype: tuple(function, function, function, function)
        """
        if initial_phases is None: initial_phases = self.initial_phases
        if distance_units != "natural" and (self.M is None or self.mu is None): raise ValueError("M and mu must be specified to convert r to physical units")
        if time_units != "natural" and (self.M is None or self.mu is None): raise ValueError("M and mu must be specified to convert t to physical units")
        
        a, p, e, x = self.a, self.p, self.e, self.x
        upsilon_r, upsilon_theta, upsilon_phi, gamma = self.upsilon_r, self.upsilon_theta, self.upsilon_phi, self.gamma
        constants = self.constants_of_motion()
        radial_roots = _radial_roots(a,p,e,constants)
        polar_roots = _polar_roots(a,x,constants)

        r_phases, t_r, phi_r = radial_solutions(a,constants,radial_roots)
        theta_phases, t_theta, phi_theta = polar_solutions(a,constants,polar_roots)
        q_t0, q_r0, q_theta0, q_phi0 = initial_phases

        distance_conversion_func = {"natural": lambda x,M: x, "mks": distance_in_meters, "cgs": distance_in_cm, "au": distance_in_au,"km": distance_in_km}
        time_conversion_func = {"natural": lambda x,M: x, "mks": time_in_seconds, "cgs": time_in_seconds, "days": time_in_days}

        # Calculate normalization constants so that t = 0 and phi = 0 at lambda = 0 when q_t0 = 0 and q_phi0 = 0 
        C_t = t_r(q_r0)+t_theta(q_theta0)
        C_phi= phi_r(q_r0)+phi_theta(q_theta0)

        def t(mino_time):
            # equation 6
            return time_conversion_func[time_units](q_t0 + gamma*mino_time + t_r(upsilon_r*mino_time+q_r0) + t_theta(upsilon_theta*mino_time+q_theta0) - C_t, self.M)
        
        def r(mino_time):
            return distance_conversion_func[distance_units](r_phases(upsilon_r*mino_time+q_r0),self.M)
        
        def theta(mino_time):
            return theta_phases(upsilon_theta*mino_time+q_theta0)
        
        def phi(mino_time):
            # equation 6
            return q_phi0 + upsilon_phi*mino_time + phi_r(upsilon_r*mino_time+q_r0) + phi_theta(upsilon_theta*mino_time+q_theta0) - C_phi
        
        return t, r, theta, phi

    